"""
    Sagemaker Engine Helpers module used to manage
    and get information on client subscription
"""
import json
from sdc_helpers.query_helper import QueryHelper

class SagemakerEngineHelpers(QueryHelper):
    """Extension on the QueryHelper"""
    def __init__(self, **kwargs):
        # instantiate the same super
        super().__init__(**kwargs)

    @staticmethod
    def build_cache_key(*, service, obj_group, prefix: list = []):
        """
            Builds a key for use with redis cache

            args:
                service (str):
                group_name (str):
            *prefix:
                any number of non-keyword arguments given by the user

            example:
                >> cache_key = SagemakerEngineHelpers.build_cache_key(
                    'recommend', 'lookup', 'prefix1', 'prefix2', 'prefix3')
                >> print(cache_key)
                'recommend-prefix1-prefix2-prefix3-lookup'
        """
        # join prefixes
        if len(prefix) > 0:
            prefixes = "-".join(prefix)

        # build cache key
        cache_key = '{service}-{prefixes}-{obj_group}'.format(
            service=service,
            prefixes=prefixes,
            obj_group=obj_group)

        return cache_key.replace(" ", "-").lower()

    # add overriding functionality
    def get_engines(self, *, client_id: str):
        """
            get engines json, assume reuse

            note:
                - single point of failure with mysql json_extract
                - this could be made even more efficiency\
                    if datasets, campaigns each when to get it their selves
                - but potential for breakages due to string query bugs
        """
        cursor = self.db_conn.cursor()

        sql = (
            'SELECT JSON_EXTRACT(properties, "$.engines") as engines'
            ' FROM `subscriptions` '
            'WHERE `subscriptions`.`client_id` = %s'
                )

        cursor.execute(sql, (client_id, ))

        try:
            engines = cursor.fetchall()[0].get('engines')
            engines = json.loads(engines)

        except ValueError:
            raise Exception(
                        ("ClientError: No Engines found"
                        "for client_id= '{}'".format(client_id))
                    )

        return engines

    def get_engine(self, *, client_id: str, slug: str):
        """
            Get a specific engine from subscription properties
            args:
                properties (dict): Subscription properties for a client
                slug (str): The engine's slug
            returns:
                engine_dict (dict): The requested engine dictionary from the properties if found
        """
        engines = self.get_engines(client_id=client_id)
        for index, engine in enumerate(engines):
            if engine.get('slug') == slug:
                return engine, index

        raise IndexError(
                    ("ClientError: No Engine matching '{}' "
                    "found for the client = '{}'".format(slug, client_id))
                )

    def get_datasets(self, *, client_id: str, engine: str):
        """Get list of all datasets for the engine"""
        # ignore index
        engine, _ = self.get_engine(client_id=client_id, slug=engine)
        # unlikely that datasets = empty([]), but not failing condition
        return engine.get('datasets', [])

    def get_dataset(self, *, client_id: str, engine: str, label: str):
        """
            Get a specific dataset from an engine dictionary
            args:
                engine (dict): Subscription properties engine
                label (str): Label describing what the dataset is for
            returns:
                dataset_dict (dict): The requested dataset dictionary from the engine if found
        """
        datasets = self.get_datasets(client_id=client_id, engine=engine)

        if datasets is not None:
            for dataset in datasets:
                # only return the datasets if the label matches
                if dataset.get('label') == label:
                    return dataset

        raise IndexError(
                        ("ClientError: No Dataset matching '{}' "
                        "found in Engine = '{}' "
                        "for client = '{}'".format(label, engine, client_id))
                )

    def get_item_from_lookup(self, *, client_id: str, engine: str, label: str, key: str):
        """
            Get a specific dataset from an engine dictionary
            args:
                engine (dict): Subscription properties engine
                label (str): Label describing what the dataset is for
            returns:
                dataset_dict (dict): The requested dataset dictionary from the engine if found
        """
        dataset = self.get_dataset(client_id=client_id, engine=engine, label=label)

        return dataset.get("data", {}).get(str(key))

    def get_campaigns(self, *, client_id: str, engine: str):
        """Get list of all datasets for the engine"""
        # ignore index
        engine, _ = self.get_engine(client_id=client_id, slug=engine)
        # unlikely that datasets = empty([]), but not failing condition
        return engine.get('campaigns', [])

    def get_campaign(self, client_id: str, engine: str, recipe: str):
        """
            Get a specific campaign from an engine dictionary
            args:
                engine (dict): Subscription properties engine
                recipe (str): The recipe of the campaign e.g related_items
                event_type (str): Supplied if the campaign only involves certain events e.g
                                only listing views, but not listing enquiries
            returns:
                campaign_dict (dict): The requested campaign dictionary from the engine if found
        """

        # get campaigns
        campaigns = self.get_campaigns(client_id=client_id, engine=engine)
        # get campaign dict
        if campaigns is not None:
            for campaign in campaigns:
                # only return the campaign if the recipe matches
                if campaign.get('recipe') == recipe:
                    return campaign

        raise IndexError(
                    ("ClientError: No Campaign matching '{}' "
                    "found for Engine = '{}' "
                    "And client = '{}' ".format(recipe, engine, client_id))
                )



"""
    Helpers module used by all engines
"""

def get_engine(*, properties: dict, slug: str, return_index: bool = False):
    """
        Get a specific engine from subscription properties

        args:
            properties (dict): Subscription properties for a client
            slug (str): The engine's slug
            return_index (bool): Return the index the engine was found in the properties

        returns:
            engine_dict (dict): The requested engine dictionary from the properties if found

    """
    engines = properties.get('engines', None)

    if engines is not None:
        for index, engine in enumerate(engines):
            if engine.get('slug') == slug:
                if return_index:
                    return engine, index

                return engine

    if return_index:
        return None, None

    return None


def get_dataset(*, engine: dict, label: str):
    """
        Get a specific dataset from an engine dictionary

        args:
            engine (dict): Subscription properties engine
            label (str): Label describing what the dataset is for

        returns:
            dataset_dict (dict): The requested dataset dictionary from the engine if found

    """
    datasets = engine.get('datasets', None)

    if datasets is not None:
        for dataset in datasets:
            if dataset.get('label') == label:
                return dataset

    return None

def get_campaign(engine: dict, **kwargs):
    """
        Get a specific campaign from an engine dictionary

        args:
            engine (dict): Subscription properties engine
            recipe (str): The recipe of the campaign e.g related_items
            event_type (str): Supplied if the campaign only involves certain events e.g
                              only listing views, but not listing enquiries

        returns:
            campaign_dict (dict): The requested campaign dictionary from the engine if found

    """
    arn = kwargs.get('arn', None)
    recipe = kwargs.get('recipe', None)
    event_type = kwargs.get('event_type', None)

    campaigns = engine.get('campaigns', None)

    if campaigns is not None:
        for campaign in campaigns:
            if (
                    (
                        arn is not None and campaign.get('arn') == arn
                    ) or
                    (
                        campaign.get('recipe') == recipe and
                        (event_type is None or campaign.get('event_type') == event_type)
                    )
            ):
                return campaign

    return None



"""
    Personalize event manager module
"""
import json
import time
from sdc_helpers.query_helper import QueryHelper
from sdc_helpers.redis_helper import RedisHelper
import boto3
import sdc_engine_helpers.helpers as helpers


class EventManager:
    """
        Manage real-time events to AWS personalize for a given set of parameters
    """
    service = None
    query_helper = None
    redis_helper = None
    parameters = None

    def __init__(self, **kwargs):
        self.query_helper = QueryHelper()
        self.redis_helper = RedisHelper()
        self.service = self.query_helper.get_service(slug='recommend')
        self.parameters = kwargs

    def __del__(self):
        del self.query_helper
        del self.redis_helper

    def get_client(self, api_key_id: str):
        """
            Determine the client with the supplied parameters

            returns:
                client (dict): The determined client

        """
        client = None

        client_id = self.parameters.get('client_id', None)
        if client_id is not None:
            client = self.query_helper.get_client(client_id=client_id)

        api_key_id = self.parameters.get('api_key_id', None)
        if api_key_id is not None:
            client = self.query_helper.get_client(api_key_id=api_key_id)

        if client is None:
            raise Exception('ClientError: Could not determine client for this request')

        return client

    def get_tracking_id(
            self,
            *,
            client_id: int
    ):
        """
            Get tracking for the given client and service

            args:
                client_id (int): The client id

            returns:
                tracking_id (string): The determined tracking id

        """
        tracking_id = None
        subscription_properties = self.query_helper.get_subscription_properties(
            client_id=client_id,
            service_id=self.service['id']
        )

        if subscription_properties:
            engine = helpers.get_engine(
                properties=subscription_properties,
                slug='personalize'
            )

            if engine is not None:
                dataset = helpers.get_dataset(
                    engine=engine,
                    label='tracking'
                )

                tracking_id = dataset.get('identifier', None)

        if not tracking_id:
            raise Exception(
                'ServerError: Could not determine tracking_id for this client'
            )

        return tracking_id

    def track_event(self):
        """
            Track the personalise event from the event
            dictionary provided by lambda

            returns:
                results (dict): A dict for logging the event

        """
        client = self.get_client(
            api_key_id=self.parameters.get('api_key_id')
        )

        event_type = self.parameters.get('eventType', 'listingView')

        tracking_id = self.get_tracking_id(
            client_id=client['id']
        )

        item_id = self.parameters.get('itemId')
        user_id = self.parameters.get('userId')
        timestamp = self.parameters.get('timestamp', int(time.time()))

        try:
            personalize_events = boto3.client('personalize-events')

            personalize_events.put_events(
                trackingId=tracking_id,
                userId=user_id,
                sessionId=user_id,
                eventList=[
                    {
                        'sentAt': timestamp,
                        'eventType': event_type,
                        'properties': json.dumps(
                            {
                                'itemId': item_id,
                            }
                        )
                    }
                ]
            )
        except Exception as ex:
            raise Exception(
                'ServerError: {exception}'.format(exception=str(ex))
            )

        print('Successful put event')

        results = {
            'client': client['name'],
            'context': self.parameters.get('context', None),
            'user_id': user_id,
            'item_id': item_id
        }
        print(results)

        return results



"""
    Personalize recommendations manager module
"""
import json
import boto3
import sdc_helpers.utils as utils
from sdc_helpers.query_helper import QueryHelper
from sdc_helpers.redis_helper import RedisHelper
import sdc_engine_helpers.helpers as helpers

class RecommendationsManager:
    """
        Manages retrieval of Personalize recommendations for a given set of parameters
    """
    service = None
    query_helper = None
    redis_helper = None
    parameters = None

    def __init__(self, **kwargs):
        self.query_helper = QueryHelper()
        self.redis_helper = RedisHelper()
        self.service = self.query_helper.get_service(slug='recommend')
        self.parameters = kwargs

    def __del__(self):
        del self.query_helper
        del self.redis_helper

    def get_recommendations(self):
        """
            Get the recommendations for the specified parameters

            returns:
                results (dict): Results from the Personalize invocation
        """
        client = self.get_client()

        item_id = self.parameters.get('itemId')

        if not item_id:
            raise Exception('ClientError: `itemId` is required')

        cache_key = 'recommend-{client_id}-{item_id}'.format(
            client_id=client['id'],
            item_id=item_id
        )

        results = self.redis_helper.redis_get(key=cache_key)

        if not results:
            campaign = self.get_campaign(client)

            if not campaign:
                raise Exception('ServerError: Could not determine campaign for this client')

            campaign_arn = campaign['arn']
            default_results = campaign.get('default_results', [])

            results = self.get_results(
                client=client,
                campaign_arn=campaign_arn,
                default_results=default_results
            )

            self.redis_helper.redis_set(
                key=cache_key,
                value=json.dumps(results),
                expiry=7200
            )
        else:
            results = json.loads(results)

        domain = utils.dict_query(
            dictionary=self.parameters,
            path='context.domain'
        )

        session_hash = utils.dict_query(
            dictionary=self.parameters,
            path='context.sessionHash'
        )

        trigger_log = {
            'trigger': {
                'client': client['name'],
                'type': 'recommendations',
                'data': {
                    'domain': domain,
                    'item_id': item_id,
                    'session_hash': session_hash,
                    'results': results
                }
            }
        }

        print(json.dumps(trigger_log))

        return results

    def get_client(self) -> dict:
        """
            Determine the client with the supplied parameters

            returns:
                client (dict): The determined client

        """
        client = None

        client_id = self.parameters.get('client_id', None)
        if client_id is not None:
            client = self.query_helper.get_client(client_id=client_id)

        api_key_id = self.parameters.get('api_key_id', None)
        if api_key_id is not None:
            client = self.query_helper.get_client(api_key_id=api_key_id)

        if client is None:
            raise Exception('ClientError: Could not determine client for this request')

        return client

    def get_campaign(self, client: dict):
        """
            Determine the campaign with the supplied parameters

            args:
                client (dict): The determined client

            returns:
                campaign (dict): Campaign from the database if found

        """
        recipe = self.parameters.get('recipe', 'related_items')
        event_type = self.parameters.get('eventType', 'all')

        subscription_properties = self.query_helper.get_subscription_properties(
            client_id=client['id'],
            service_id=self.service['id']
        )

        if subscription_properties is not None:
            engine = helpers.get_engine(
                properties=subscription_properties,
                slug='personalize'
            )

            if engine is not None:
                return helpers.get_campaign(
                    engine=engine,
                    recipe=recipe,
                    event_type=event_type
                )

        return None

    def get_results(self, *, client: dict, campaign_arn: str, default_results: list) -> list:
        """
            Determine the results for this request

            args:
                client (object): The client requesting the recommendation
                campaign_arn (str): The AWS Personalize campaign ARN
                default_results (list): The campaign's default results. If the recommendations
                                        are similar to the default results, an empty results
                                        list is returned as the default results are returned
                                        for an an item that does not exist in the solution the
                                        campaign is referencing

            returns:
                results (list): Results from the Personalize invocation

        """
        item_id = self.parameters.get('itemId')
        num_results = self.parameters.get('numResults', 25)

        results = []

        try:
            personalize_runtime = boto3.client('personalize-runtime')

            response = personalize_runtime.get_recommendations(
                campaignArn=campaign_arn,
                itemId=item_id,
                numResults=num_results
            )
        except Exception as ex:
            raise Exception('ServerError: {exception}'.format(exception=str(ex)))

        for item in response['itemList']:
            results.append(item['itemId'])

        # Check if the results are the same or similar to the default results
        # and return empty array if so
        intersection = set(default_results).intersection(results)

        if (
                len(results) == len(intersection) or
                len(intersection) / len(results) >= 0.8
        ):
            if len(results) != len(intersection):
                print(
                    'Results for {client_name} are similar but not the same to the default results '
                    'which probably means default results are out of sync. Saving default results'
                    .format(client_name=client['name'])
                )

            results = []

        return results



"""
    Sagemaker recommendations manager module
"""
import json
import boto3
import sdc_helpers.utils as utils
from sdc_helpers.redis_helper import RedisHelper
from sdc_engine_helpers.sagemaker.sagemaker_query_helpers import \
    SagemakerEngineHelpers

class RecommendationsManager:
    """
        Manages retrieval of Sagemaker recommendations for a given set of parameters
    """
    client = {}
    engine = None

    def __init__(self, **kwargs):
        """
            kwargs:
                service_slug (str): slug of service to use
                rds_config (dict): config for mysql database
                redis_config (dict): config for redis cache database
                item_ids (str): name of item id lookup in mysql db
                item_codes (str): name of item code lookup in mysql db
        """
        # get rds config
        rds_config = kwargs.get('rds_config', {})
        if not isinstance(rds_config, dict):
            raise TypeError('Only dict is supported for rds_config')

        # get redis config
        redis_config = kwargs.get('redis_config', {})
        if not isinstance(redis_config, dict):
            raise TypeError('Only dict is supported for redis_config')

        # instantiate query helpers
        self.query_helper = SagemakerEngineHelpers(
            rds_config=rds_config,
            redis_config=redis_config)

        self.redis_helper = RedisHelper(**redis_config)
        # service slug is the same for all engines
        service_slug = kwargs.pop('service_slug', 'recommend')

        # get name of lookups
        self.item_id_identifier = kwargs.pop('item_id_identifier', 'item_ids')
        self.item_code_identifier = kwargs.pop('item_code_identifier', 'item_codes')

        # load service details from service table
        self.service = self.query_helper.get_service(slug=service_slug)

    def __del__(self):
        """
            Close connections on delete
        """
        del self.query_helper
        del self.redis_helper

    def get_recommendations(
        self, *,
        client_id: str,
        item_id: str, \
        engine: str,
        context: dict = {}, **kwargs):
        """
            Get the recommendations for the specified parameters

            parameters:
                - dictionary of parameter arguments
            returns:
                results (dict): Results from the Sagemaker invocation
        """
        # get client metadata from client_id
        self.client = self.get_client(client_id=client_id)
        self.engine = engine
        # recipe is set by client config
        recipe = kwargs.get('recipe', 'related_items')

        campaign = self.get_campaign(
            client_id=self.client['id'],
            engine=self.engine,
            recipe=recipe)

        if not campaign:
            raise Exception(
                            ('ServerError: Could not determine'
                            ' campaign for this client')
                        )

        # changed campaign_arn to be campaign_name
        endpoint_name = campaign['arn']

        results = self.get_results(
            item_id=item_id,
            endpoint_name=endpoint_name,
            num_results=10)

        domain = utils.dict_query(
            dictionary=context,
            path='domain')

        session_hash = utils.dict_query(
            dictionary=context,
            path='sessionHash')

        # trigger will be written with alphabetical order in mind
        trigger_log = {
                'trigger': {
                        'client': self.client['name'],
                        'type': 'recommendations',
                        'data': {
                            'domain': domain,
                            'item_id': item_id,
                            'session_hash': session_hash,
                            'results': results
                            }
                        }
                    }

        # write to lambda output logs
        print(json.dumps(trigger_log))
        return results

    def get_client(
        self, *,
        client_id: str = None,
        api_key_id: str = None) -> dict:
        """
            Determine the client with the supplied parameters
            Expects either client_id or api_key_id
            args:
                client_id (int): client id to fetch
                api_key_id (int): api_key_id to fetch client_id
            returns:
                client (dict): The determined client

        """
        # Possible double retrieve
        # this is potentially costly
        if client_id is not None:
            # do this first if client_id no None
            client = self.query_helper.get_client(client_id=client_id)
        elif api_key_id is not None:
            # do this if client_id==None
            client = self.query_helper.get_client(api_key_id=api_key_id)
        if client is None:
            raise Exception(
                    ('ClientError: Could not determine '
                    'client for this request')
                    )
        return client

    def get_campaign(
        self, *,
        client_id: str,
        engine: str,
        recipe='related_items') -> dict:
        """
            Determine the campaign with the supplied parameters
            note:
                - Currently: Campaign arn = (Sagemaker Engpoint Name)

            args:
                client_id (int): client id
                engine (str): engine slug
                recipe (str): recipe name

            returns:
                results (dict): Campaign from the database

        """
        campaign = self.query_helper.get_campaign(
            client_id=client_id,
            engine=engine,
            recipe=recipe)

        return campaign

    def encode_item_code(
        self, *,
        item_ids: list,
        from_cache: bool = True):
        """
            Lookup a list of item codes

            args:
                 item_ids (list) : list of item id's
                 from_cache (bool) : Get lookup from cache first
            return:
                item_codes (list) : list of item codes
        """

        item_codes = []

        cache_key = self.query_helper.build_cache_key(
            service=self.service['name'],
            obj_group=self.item_code_identifier,
            prefix=[self.engine, self.client['name']])

        if from_cache:
            # encode item_id -> item_code
            item_codes = self.redis_helper.redis_hmget(
                hashkey=cache_key,
                keys=item_ids)

        for i, item_code in enumerate(item_codes):
            # decode item_ids bytes -> str
            if isinstance(item_code, bytes):
                item_code = item_code.decode("utf-8")

            # check if item_id is none
            if item_code is None:
                # get item_id from database if none
                key = item_ids[i]
                item_code = self.query_helper.get_item_from_lookup(
                    client_id=self.client['id'],
                    engine=self.engine,
                    label=self.item_code_identifier,
                    key=key)

                if item_code is not None:
                    # store in cache
                    self.redis_helper.redis_hmset(hashkey=cache_key, dict_obj={key:item_code})

            item_codes[i] = item_code
        return item_codes

    def decode_item_code(
        self, *,
        item_codes: list,
        from_cache: bool = True):
        """"
            Lookup a list of item codes

            args:
                 item_codes (list) : list of item codes
                 from_cache (bool) : Get lookup from cache first
            return:
                item_ids (list) : list of item id
        """
        item_ids = []

        cache_key = self.query_helper.build_cache_key(
            service=self.service['name'],
            obj_group=self.item_id_identifier,
            prefix=[self.engine, self.client['name']])

        if from_cache:
            # decode item_codes -> item_ids
            item_ids = self.redis_helper.redis_hmget(
                hashkey=cache_key,
                keys=item_codes)
            
        for i, item_id in enumerate(item_ids):
            if isinstance(item_id, bytes):
                # decode item_ids bytes -> str
                item_id = item_id.decode("utf-8")

            # check if item_id is none
            if item_id is None:
                # get item_id from database if none
                key = item_codes[i]
                item_id = self.query_helper.get_item_from_lookup(
                    client_id=self.client['id'],
                    engine=self.engine,
                    label=self.item_id_identifier,
                    key=key)

                if item_id is not None:
                    # store in cache
                    self.redis_helper.redis_hmset(hashkey=cache_key, dict_obj={key:item_id})

            item_ids[i] = item_id

        return item_ids

    def get_results(
        self, *,
        item_id: str,
        endpoint_name: str,
        num_results: int = 10) -> list:
        """
            Determine the results for this request

            args:
                item_id (str): The client requesting the recommendation
                endpoint_name (str): The AWS Sagemaker endpoint_name
                num_results (int) : Number of recommendation results to return

            returns:
                results (list): Sagemaker invocation results

        """
        # lookup item code
        item_code = self.encode_item_code(item_ids=[item_id])
        # test that item_code is valid
        if (item_code == list()) | (item_code[0] is None):
            raise Exception('ClientError: item_id was not found')

        try:
            sagemaker_runtime = boto3.client('sagemaker-runtime')
            # make the recommender ignore self recommendation
            event = {
                'item_code':int(item_code[0]),
                'args':{
                    'n_similar':num_results,
                    'exclude':[int(item_code[0])]}}
            payload = json.dumps(event)

            # get recommendations
            response = sagemaker_runtime.invoke_endpoint(
                EndpointName=endpoint_name,
                ContentType='application/json',
                Body=payload)

            # decode application/json response
            response_decoded = json.loads(response['Body'].read().decode())

            #extract itemcodes
            recommendations = [r['item_code'] for r in response_decoded]

        except sagemaker_runtime.exceptions.ModelError:
            # insure model errors are returned as 500
            msg = "ServerError: 500 Internal Server Error"
            raise Exception(msg)

        # decode recommendations bytes -> str
        results = self.decode_item_code(item_codes=recommendations)
        # check list of item_ids are not all None
        if all(item_id is None for item_id in results):
            raise Exception(
                            ('ClientError: Unable to decode'
                            ' recommendations to item_id')
                        )

        return results
